/* 
 * File:   mm.h
 * Author: wicker
 *
 * Created on 8 juillet 2011, 16:31
 */

#include <vector>
#include <iostream>
#include <cmath>
#include <cfloat>
#include <string>
using namespace std;

#include "Lawc.h"
#include "Gaussian.h"
#include "SGaussian.h"
#include "SStudent.h"
#include "FisherVonMises.h"
#include "Dirichlet.h"
#include "PointsSet.h"
#ifndef MM_H
#define	MM_H

#define LAW_TYPE_GAUSSIAN 1
using namespace std;

template <typename Law> class MM {
public:
    int nbComponents;
    bool noiseComponent;
    bool noiseComponentEmpty;
    double logDensityNoise;
    double *taus;
    Law **components;
    double logLikelihood;

    MM();
    MM(string criterium, PointsSet *pointsSet, double proba, int maxNbComponents, bool noiseComponent);
    MM(int nbComponents, PointsSet *pointsSet, bool noiseComponent) throw (string);
    void maximisation(PointsSet *pointsSet) throw (string);
    void expectation(PointsSet *pointsSet);
    int selectionForAIC(PointsSet *pointsSet, int nbComponentsMax, MM<Law> **allSolutions);
    int selectionForBIC(PointsSet *pointsSet, int nbComponentsMax, MM<Law> **allSolutions);
    int selectionForBUCwithGivenP(int bestSolution, PointsSet *pointsSet, int nbComponentsMax, MM<Law> **allSolutions, double proba);
    int selectionForBUC(PointsSet *pointsSet, int nbComponentsMax, MM<Law> **allSolutions, double proba);
    void print();
private:
    int nbIterations;
    bool diag;
};

template <typename Law> MM<Law>::MM() {

}

/**********************************************************/
/*                                                        */
/*Fit a mixture model with nbComponents on a set of points*/
/*                                                        */

/**********************************************************/

template <typename Law> MM<Law>::MM(int nbComponents, PointsSet *pointsSet, bool noiseComponent) throw (string) {
    int completedNbComponents;
    this->noiseComponent = noiseComponent;
    this->noiseComponentEmpty=true;
    this->nbComponents = nbComponents;

    if (this->noiseComponent == true) {
        completedNbComponents = this->nbComponents + 1;
    } else {
        completedNbComponents=this->nbComponents;
    }

    cout << "nbComponents " << this->nbComponents << endl;
    this->nbIterations = 30;
    if (pointsSet->dimension <= 20) {
        this->diag = false;
    } else {
        this->diag = true;
    }
    this->taus = new double[completedNbComponents];
    this->components = new Law*[completedNbComponents];
    for (int i = 0; i < completedNbComponents; i++) {
        this->components[i] = new Law(pointsSet->dimension, this->diag);
    }
    if (this->noiseComponent == true) {
        this->logDensityNoise =
                ((Lawc *) this->components[0])->computeLogDensityNoise(pointsSet->size, pointsSet->values);
    }
    //vector <Law> components(nbComponents);

    int nbTrials, maxNbTrials = 10;
    nbTrials = 0;
    while (nbTrials < maxNbTrials) {
        try {
            pointsSet->kmeans(this->nbComponents);
            break;
        } catch (string str) {
            nbTrials++;
        }
    }

    int indexCluster, indexPoint;

    for (int iteration = 0; iteration<this->nbIterations; iteration++) {
        /*MAXIMISATION*/
       // cout << iteration << endl;
        try {
            //  cout << "maximisation " << endl;
            this->maximisation(pointsSet);
        } catch (string str) {
            throw str;
        }
        /*EXPECTATION*/
        //  cout << "expectation " << endl;

        this->expectation(pointsSet);
        //cout << "loglikelihood : "<< this->logLikelihood << endl;
    }
    if((this->noiseComponent==true)&&(this->taus[nbComponents]>DBL_MIN)) {
        this->noiseComponentEmpty=false;
    }
    /*for (int i = 0; i < nbComponents; i++) {
        ((Lawc *)this->components[i])->print();
    }*/

}

template <typename Law> void MM<Law>::maximisation(PointsSet *pointsSet) throw (string) {
    int completedNbComponents;

    PointsSet *clusters;
    if (this->noiseComponent == true) {
        completedNbComponents = nbComponents + 1;
    } else {
        completedNbComponents=nbComponents;
    }
    clusters = new PointsSet[completedNbComponents];
    int indexCluster, indexPoint;
    for (indexCluster = 0; indexCluster < completedNbComponents; indexCluster++) {
        clusters[indexCluster].size = 0;
    }
    for (indexPoint = 0; indexPoint < pointsSet->size; indexPoint++) {
        indexCluster = pointsSet->clusters[indexPoint];
        clusters[indexCluster].size++;
    }

    for (indexCluster = 0; indexCluster < completedNbComponents; indexCluster++) {
        this->taus[indexCluster] = (double) clusters[indexCluster].size / (double) pointsSet->size;

        if ((this->diag == false) && (clusters[indexCluster].size < pointsSet->dimension + 1)&&(indexCluster<nbComponents)) {
            throw string("empty cluster");
        }

        clusters[indexCluster].values = new double*[pointsSet->dimension];
        for (int i = 0; i < pointsSet->dimension; i++) {
            clusters[indexCluster].values[i] = new double[clusters[indexCluster].size];
        }
        clusters[indexCluster].size = 0;
    }
    for (indexPoint = 0; indexPoint < pointsSet->size; indexPoint++) {
        indexCluster = pointsSet->clusters[indexPoint];
        for (int i = 0; i < pointsSet->dimension; i++) {
            clusters[indexCluster].values[i][clusters[indexCluster].size] = pointsSet->values[i][indexPoint];
        }
        clusters[indexCluster].size++;
    }


    for (indexCluster = 0; indexCluster < this->nbComponents; indexCluster++) {
        ((Lawc *) this->components[indexCluster])->estimation(clusters[indexCluster].size, clusters[indexCluster].values);
    }

}

template <typename Law> void MM<Law>::expectation(PointsSet *pointsSet) {
    int indexPoint, indexCluster;
    int bestCluster, bestLikelihood;
    double *values, likelihood;
    values = new double[pointsSet->dimension];
    int j;
    this->logLikelihood = 0;

    for (indexPoint = 0; indexPoint < pointsSet->size; indexPoint++) {
        likelihood = 0;
        for (j = 0; j < pointsSet->dimension; j++) {
            values[j] = pointsSet->values[j][indexPoint];
        }

        int bestCluster;
        double logl, bestlogl;
        bestlogl = -DBL_MAX;
        for (indexCluster = 0; indexCluster < nbComponents; indexCluster++) {
            logl = ((Lawc *) components[indexCluster])->computeLogDensity(values) +
                    log(this->taus[indexCluster]);
            if (logl > bestlogl) {
                bestlogl = logl;
                pointsSet->clusters[indexPoint] = indexCluster;
            }
            likelihood += exp(logl);
        }
        if (noiseComponent == true) {
            if(this->taus[nbComponents]<DBL_MIN) {
                this->taus[nbComponents]=0.01;
            }
            logl = this->logDensityNoise + log(this->taus[nbComponents]);
            if (logl > bestlogl) {
                bestlogl = logl;
                pointsSet->clusters[indexPoint] = nbComponents;
            }
            likelihood += exp(logl);
        }
        this->logLikelihood += log(likelihood);

    }
    delete []values;
}

/******************************************************************/
/*                                                                */
/*Fit a mixture model with unknown nbComponents on a set of points*/
/*                                                                */

/******************************************************************/

template <typename Law> MM<Law>::MM(string criterium, PointsSet *pointsSet, double proba, int maxNbComponents, bool noiseComponent) {
    int nbComponents;
    this->noiseComponent = noiseComponent;

    if (maxNbComponents == -1) {
        maxNbComponents = (int) sqrt((double) pointsSet->size);
    }
    MM<Law> **allSolutions = new MM<Law>*[maxNbComponents];
    for (nbComponents = 1; nbComponents <= maxNbComponents; nbComponents++) {
        try {
            allSolutions[nbComponents - 1] = new MM<Law > (nbComponents, pointsSet, this->noiseComponent);
        } catch (string str) {
            allSolutions[nbComponents - 1] = new MM<Law > ();
            allSolutions[nbComponents - 1]->logLikelihood = -DBL_MAX;
            //  break;
        }
        cout << nbComponents << " " << allSolutions[nbComponents - 1]->logLikelihood << endl;
    }
    /*for (int i = nbComponents - 1; i < maxNbComponents; i++) {
        allSolutions[i] = new MM<Law > ();
        allSolutions[i]->logLikelihood = -DBL_MAX;
    }*/

    /*estimate all solutions*/
    int bestSolution;
    if ((criterium.compare("bic") == 0) || (criterium.compare("aic") == 0) ||
            (criterium.compare("ibic") == 0) || (criterium.compare("iaic") == 0)) {
        if ((criterium.compare("aic") == 0) || (criterium.compare("iaic") == 0)) {
            bestSolution = selectionForAIC(pointsSet, maxNbComponents, allSolutions);
        } else {
            bestSolution = selectionForBIC(pointsSet, maxNbComponents, allSolutions);
        }

        int p;
        if ((criterium.compare("iaic") == 0) || (criterium.compare("ibic") == 0)) {
            bestSolution = selectionForBUCwithGivenP(bestSolution, pointsSet, maxNbComponents, allSolutions, proba);
        }
    } else {
        bestSolution = selectionForBUC(pointsSet, maxNbComponents, allSolutions, proba);
    }
    /*cout << "bestsolution " << bestSolution << endl;
    for (int i = 0; i < bestSolution; i++) {
        cout << allSolutions[bestSolution - 1]->taus[i] << endl;
    }*/
    this->nbComponents = bestSolution;
    this->taus = allSolutions[bestSolution - 1]->taus;
    this->components = allSolutions[bestSolution - 1]->components;
    this->logLikelihood = allSolutions[bestSolution - 1]->logLikelihood;
    this->noiseComponentEmpty=allSolutions[bestSolution-1]->noiseComponentEmpty;
}

template <typename Law> int MM<Law>::selectionForAIC(PointsSet *pointsSet, int nbComponentsMax,
        MM<Law> **allSolutions) {
    int bestSolution, nbComponents;
    double penalty;
    double *criteria = new double[nbComponentsMax];

    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        //cout << "nbComponents " << nbComponents << endl;
        penalty = (double) (nbComponents - 1)+(double) nbComponents *
                (double) (allSolutions[0]->components[0]->complexity());

        criteria[nbComponents - 1] = allSolutions[nbComponents - 1]->logLikelihood - penalty;
    }
    double bestCriterium = -DBL_MAX;
    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        if (criteria[nbComponents - 1] > bestCriterium) {
            bestCriterium = criteria[nbComponents - 1];
            bestSolution = nbComponents;
        }
    }
    delete []criteria;
    return bestSolution;
}

template <typename Law> int MM<Law>::selectionForBIC(PointsSet *pointsSet, int nbComponentsMax,
        MM<Law> **allSolutions) {
    int bestSolution, nbComponents;
    double penalty;
    double *criteria = new double[nbComponentsMax];

    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        //cout << "nbComponents " << nbComponents << endl;
        penalty = (double) (nbComponents - 1)+(double) nbComponents *
                (double) (allSolutions[0]->components[0]->complexity());

        criteria[nbComponents - 1] = allSolutions[nbComponents - 1]->logLikelihood - 0.5 * log((double) pointsSet->size) * penalty;
    }
    double bestCriterium = -DBL_MAX;
    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        if (criteria[nbComponents - 1] > bestCriterium) {
            bestCriterium = criteria[nbComponents - 1];
            bestSolution = nbComponents;
        }
    }
    delete []criteria;
    return bestSolution;
}

template <typename Law> int MM<Law>::selectionForBUCwithGivenP(int bestSolution, PointsSet *pointsSet,
        int nbComponentsMax, MM<Law> **allSolutions, double proba) {
    int nbComponents;
    double x;
    double *criteria = new double[nbComponentsMax];

    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        x = -log(proba);
        criteria[nbComponents - 1] = allSolutions[nbComponents - 1]->logLikelihood -
                ((double) (sqrt((double) pointsSet->size)*(double) pointsSet->dimension * nbComponents)) / ((double) bestSolution) * x / 3.0;
    }
    double bestCriterium = -DBL_MAX;
    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        if (criteria[nbComponents - 1] > bestCriterium) {
            bestCriterium = criteria[nbComponents - 1];
            bestSolution = nbComponents;
        }
    }
    delete []criteria;
    return bestSolution;
}

template <typename Law> int MM<Law>::selectionForBUC(PointsSet *pointsSet,
        int nbComponentsMax, MM<Law> **allSolutions, double proba) {
    int computedNbComponents, nbComponents, bestSolution;
    double chiSquare, bestChiSquare = DBL_MAX;
    double temp;

    int lastPossibleNumberOfComponents = nbComponentsMax - 1;

    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        cout << allSolutions[nbComponents - 1]->logLikelihood << endl;
        if (allSolutions[nbComponents - 1]->logLikelihood > -DBL_MAX) {
            lastPossibleNumberOfComponents = nbComponents;
        }
    }
    int smallestComputedNbComponents;
    for (nbComponents = 1; nbComponents <= nbComponentsMax; nbComponents++) {
        computedNbComponents = selectionForBUCwithGivenP(nbComponents, pointsSet, nbComponentsMax, allSolutions, proba);
        if (nbComponents == 1) {
            smallestComputedNbComponents = computedNbComponents;
        }
        temp = ((double) computedNbComponents - (double) nbComponents) / (double) nbComponents;
        chiSquare = (temp * temp);
        cout << nbComponents << " " << computedNbComponents << endl;
        //chiSquare=fabs((double)nbComponents-(double)proposedNbComponents);        
        if ((bestChiSquare >= chiSquare) && (
                (computedNbComponents == smallestComputedNbComponents) ||
                (computedNbComponents < lastPossibleNumberOfComponents))) {
            bestChiSquare = chiSquare;
            bestSolution = computedNbComponents;
        }
    }
    return bestSolution;
}

#endif	/* MM_H */

